
/* file pscan_08.c
 created ven. déc. 17 10:27:30 CET 2021
 version ven. déc. 24 08:35:37 CET 2021 pscan_03.c
 version ven. déc. 24 12:06:21 CET 2021 pscan_04.c
 version ven. déc. 24 19:54:50 CET 2021 pscan_05.c
 version ven. déc. 24 21:11:43 CET 2021 pscan_06.c
 version sam. déc. 25 21:57:35 CET 2021 pscan_07.c : adding dual-lane scan
 version lun. déc. 27 09:17:48 CET 2021 pscan_08.c : refresh timer, X=0, ...

 gcc -Wall -Os -DWIDTH=16 -lpthread -D_REENTRANT -DTHREADS=4 pscan_07.c -o pscan && ./pscan 1 2 plop

*/

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <inttypes.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <pthread.h>
#include <time.h>

#ifndef THREADS
#define THREADS (4)
#endif

#ifndef REFRESH
#define REFRESH (10)
#endif
int refresh_counter;

pthread_t threads[THREADS];

#ifndef HZ
#define HZ (10) // Polling frequency in Hz
#endif

// Handle the PEAC parameters

#ifndef WIDTH
#define WIDTH (16)
#endif

#define SIZE  (1<<WIDTH)
#define TOTAL ((((SIZE+1ULL)*SIZE)>>1)-2ULL)
#define MASK  (SIZE-1)

// FIFO definitions
#ifndef FIFOLOG2
#define FIFOLOG2   (14)
#endif
#define FIFOLENGTH (1<<(FIFOLOG2))
#define FIFOMASK   ((FIFOLENGTH)-1)

// Comet Lake cache line is 64 bytes,
// take one line per FIFO entry
// and for the pointers to prevent cache thrashing
// (and reduce line sharing conflicts)
typedef struct {
  uint64_t tail, head, finished, pad[5];
} indices_t;

volatile indices_t ptr_job[THREADS],
                   ptr_res[THREADS];

uint64_t FIFO_jobs[THREADS][FIFOLENGTH];

typedef struct {
  uint64_t Xinit, Xend, steps;
} result_t;

result_t FIFO_res[THREADS][FIFOLENGTH];

#define FIFOEMPTY(FIFO,THREADNR) (FIFO[THREADNR].tail ==  FIFO[THREADNR].head)
#define FIFOFULL(FIFO,THREADNR)  (FIFO[THREADNR].tail ==((FIFO[THREADNR].head+1)&FIFOMASK))

#define FIFOPUTJOB(THREADNR,VALUE) {              \
    uint64_t head=ptr_job[THREADNR].head;         \
    FIFO_jobs[THREADNR][head] = VALUE;            \
    ptr_job[THREADNR].head = (head+1) & FIFOMASK; \
  }
//  printf(" + adding %"PRId64" to job FIFO %d at index %"PRId64"\n", VALUE, THREADNR, head);

#define FIFOGETJOB(THREADNR,VALUE) {              \
    uint64_t tail=ptr_job[THREADNR].tail;         \
    VALUE = FIFO_jobs[THREADNR][tail];            \
    ptr_job[THREADNR].tail = (tail+1) & FIFOMASK; \
  }
//  printf(" - getting %"PRId64" from job FIFO %d at index %"PRId64"\n", VALUE, (int)THREADNR, tail);


// Handle the log output format:
#include "p7ck.c"
#ifndef BUFFZIZE   // Could be reduced quite a lot for w > 21, otherwise display update is tooooooo sloooooooooow
#define BUFFZIZE (32768)
#endif

p7k_struct results[BUFFZIZE];
int fout;
int p7k_count=0;
uint64_t total_written, whole_range;

void flush_results() {
  int s=p7k_count * sizeof results[0];
  if (write(fout, &results, s) != s) {
    perror ("couldn't write results");
    exit(EXIT_FAILURE);
  }
  total_written += p7k_count;
  fprintf(stderr," %"PRId64"/%"PRId64"  \r", total_written, whole_range);
  p7k_count=0;
}

void semiarc(uint64_t Xorigin, uint64_t *len, uint64_t *Xcross) {
  uint64_t arclen=0;
  uint32_t
    X = Xorigin,
    Y = 0,
    C = 0;

    if (X==0) {
      *Xcross = 1;
      *len = 1;
      return;
    }

loopstart:
      C += X + Y;
      Y = C & MASK;  C = C >> WIDTH;
      if (X==0)
        goto endofloopodd;
      if (X==MASK)
        goto endofloopodd_comp;

      C += X + Y;
      arclen+=2;
      X = C & MASK;  C = C >> WIDTH;
      if (Y==MASK)
        goto endofloopeven_comp;
      if (Y!=0)
        goto loopstart;

//endofloopeven:
  *Xcross = X;
  *len = arclen;
  return;

endofloopeven_comp:
  *Xcross = MASK-X;
  *len = arclen;
  return;

endofloopodd:
  *Xcross = Y;
  *len = arclen+1;
  return;

endofloopodd_comp:
  *Xcross = MASK-Y;
  *len = arclen+1;
  return;
}


void *scan_job(void *arg) {
  // OK it's ugly but the arg pointer is used
  // to contain the thread # as an integer...
  long int threadnr= (long int)arg,
    iteration=0;  // this one is not strictly useful
  uint64_t Xinit, Xend, head, next, len;
  result_t *R;
  struct timespec ts={1, 0 }; // 1s
//  int used_lanes=0;
#define MAX_LANES 2

#ifdef VERBOSE
  printf("thread %ld started.\n", threadnr);
#endif
  do {
    while (!FIFOEMPTY(ptr_job, threadnr)) {
      iteration++;

/*
typedef struct {
  uint64_t
    I1, X1, Y1, C1, Start1,
    I2, X2, Y2, C2, Start2,
    time, pad;
} state_t;

state_t states[THREADS];
*/

      FIFOGETJOB(threadnr,Xinit)
      // Xinit >= 1 here.
      semiarc(Xinit, &len, &Xend);

      // Enqueue the result
      head = ptr_res[threadnr].head;
      next = (head+1)&FIFOMASK;
      while (ptr_res[threadnr].tail == next) {
        printf("RESULT FIFO %ld IS FULL, WAITING...\n", threadnr);
        nanosleep(&ts, NULL);
      }
#ifdef VERBOSE
      printf("thread %ld finished after %"PRId64" iteration with Xinit=%"PRId64", writing %"PRId64" to index %"PRId64".\n",
        threadnr, len, Xinit, Xend, head);
#endif
      // commit the result (Xinit,Xend,len) to the result FIFO
      R = &FIFO_res[threadnr][head];
      R->Xinit = Xinit;
      R->Xend  = Xend;
      R->steps = len;
      __sync_synchronize(); // ensure the buffers are flushed and data propagated
      ptr_res[threadnr].head = next; // commit previously written data.
    }
  } while(0);

  ptr_job[threadnr].finished=1;
#ifdef VERBOSE
  printf("thread %ld ended\n", threadnr);
#endif
  return NULL;
}

// And finally:

int main(int argc, char *argv[]) {
  uint64_t arclen=0;
  uint32_t Xorig, Xend;
  int64_t Xmax=SIZE, X=0, last_X;
  int active_threads, current_thread, running;
  char *c, *outname="log.p7k";
  struct timespec ts={0, 1000000000/HZ };

  p7k_count=0;

// Parse the command line arguments
  if (argc != 1) {
    if (argc == 4) {
      X=strtoul(argv[1], &c, 10);
      if (errno || *c!=0 || X > MASK || X < 0) {
        printf("invalid Xinit parameter : %s ??\n", argv[1]);
        return EXIT_FAILURE;
      }
      Xmax=strtoul(argv[2], &c, 10);
      if (errno || *c!=0 || Xmax > SIZE || Xmax <= X) {
        printf("invalid Xend parameter : %s ??\n", argv[2]);
        return EXIT_FAILURE;
      }
      outname=argv[3];
    }
    else {
      printf("Usage : %s (origin X) (end X) (output file name)\nWidth=%d bits\n", argv[0], WIDTH);
      return EXIT_FAILURE;
    }
  }

  if ((fout=open(outname, O_CREAT|O_WRONLY|O_TRUNC,
     S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH | O_RDWR)) == -1) {
    perror("can't open output file for writing");
    return EXIT_FAILURE;
  }

  printf("Scanning w%d from X=%"PRId64" (included) to %"PRId64" (excluded)\n", WIDTH, X, Xmax);
  whole_range = Xmax - X;
  total_written = 0;

  // Init & fill the input FIFOs
  bzero((void *)ptr_job,   sizeof ptr_job);
  bzero((void *)ptr_res,   sizeof ptr_res);
  bzero((void *)FIFO_jobs, sizeof FIFO_jobs);
  bzero((void *)FIFO_res,  sizeof FIFO_res);

  active_threads=0;
  current_thread=0;
  last_X = X + (THREADS*FIFOMASK);
  if (last_X > Xmax)
      last_X = Xmax;
#ifdef VERBOSE
  printf(" last_X = %"PRId64"\n", last_X);
#endif

  while (1) {
    if (X==0) {
      // For convenience and speed, I moved this special case here.
      // This shot-circuits the call, yet still allows the thread to start
      // so X=0 is taken into account and the program is correctly stopped.
      // This condition can only happen as the very first iteration.
      // FIFO_res[threadnr][0].Xinit = 0;
      FIFO_res[0][0].Xend  = 1;
      FIFO_res[0][0].steps = 1;
      ptr_res[0].head = 1;
    }
    else
      FIFOPUTJOB(current_thread,X)

    X++;
    if (X >= last_X)
      break;
    current_thread++;
    if (current_thread >= THREADS)
        current_thread=0;
    if (active_threads < current_thread)
        active_threads = current_thread;
  };

  printf("Spawning %d threads.\n", active_threads+1);
  for(current_thread=0; current_thread <= active_threads; current_thread++) {
    pthread_create( &threads[current_thread], NULL, &scan_job, (void *)(long)current_thread);
  }

  refresh_counter=0;

// Wait loop & log management
  do {
    nanosleep(&ts, NULL);
    refresh_counter++;

    running = 0;
    for (current_thread=0; current_thread <= active_threads; current_thread++) {
      uint64_t f=ptr_job[current_thread].finished;
      if (f<2) {

        // read the result FIFO
        while (!FIFOEMPTY(ptr_res, current_thread)) {
          // do something
          uint64_t tail=ptr_res[current_thread].tail;
          Xorig  = FIFO_res[current_thread][tail].Xinit;
          Xend   = FIFO_res[current_thread][tail].Xend;
          arclen = FIFO_res[current_thread][tail].steps;
#ifdef VERBOSE
          printf("  - getting %u:%"PRId64"->%u from result FIFO %d at index %"PRId64"\n",
              Xorig, arclen, Xend, current_thread, tail);
#endif
          ptr_res[current_thread].tail = (tail+1) & FIFOMASK;

          encode_p7k( Xorig, Xend, 1, arclen, &results[p7k_count]);
          p7k_count++;

          if ((p7k_count >= BUFFZIZE)
            || (refresh_counter > (HZ*REFRESH))) {
            refresh_counter=0;
            flush_results();
          }
        }

        if (f==0) {
          running++;
          // check for room of new jobs in the jobs FIFO
          while (( X < Xmax) && !FIFOFULL(ptr_job, current_thread)) {
            FIFOPUTJOB(current_thread,X)
            X++;
          }
        }
        else { // f==1 so terminate cleanly:
          ptr_job[current_thread].finished=2;
          pthread_join( threads[current_thread], NULL);
        }
      }
    }
#ifdef VERBOSE
    printf(" poll %d.\n", running);
#endif
  } while (running > 0);

  flush_results();
  printf(" %s done.\n", outname);
  return EXIT_SUCCESS;
}
